package de.mannheim.ids.sruws;

import java.io.BufferedInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.UnsupportedEncodingException;
import java.net.URI;
import java.net.URISyntaxException;
import java.net.URLDecoder;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.servlet.ServletException;
import javax.servlet.http.HttpServlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.parsers.SAXParser;
import javax.xml.parsers.SAXParserFactory;
import javax.xml.stream.XMLInputFactory;
import javax.xml.stream.XMLStreamException;
import javax.xml.stream.XMLStreamReader;
import javax.xml.stream.events.XMLEvent;

import org.apache.http.HttpEntity;
import org.apache.http.client.methods.CloseableHttpResponse;
import org.apache.http.client.methods.HttpGet;
import org.apache.http.client.utils.URIBuilder;
import org.apache.http.impl.client.CloseableHttpClient;
import org.apache.http.impl.client.HttpClients;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.xml.sax.SAXException;

import eu.clarin.weblicht.wlfxb.io.WLDObjector;
import eu.clarin.weblicht.wlfxb.io.WLFormatException;
import eu.clarin.weblicht.wlfxb.tc.api.MatchedCorpus;
import eu.clarin.weblicht.wlfxb.tc.api.MatchesLayer;
import eu.clarin.weblicht.wlfxb.tc.api.SentencesLayer;
import eu.clarin.weblicht.wlfxb.tc.api.Token;
import eu.clarin.weblicht.wlfxb.tc.api.TokensLayer;
import eu.clarin.weblicht.wlfxb.tc.xb.TextCorpusStored;
import eu.clarin.weblicht.wlfxb.xb.WLData;

/**
 * @author margaretha
 * */

public class SRU_WS extends HttpServlet{
	
	private String endpointUri;
	private String endpointBase = "de.mannheim.ids.sruws.endpoint.";
	private String DUMMY_PID = "Wikipedia";
	
	private static final String HITS_NS = "http://clarin.eu/fcs/dataview/hits";
    private static final String HITS_PREFIX = "hits";    
    private static final String FCS_NS = "http://clarin.eu/fcs/1.0";
    private static final String FCS_PREFIX = "fcs";    
    //private static final String FCS_KWIC_NS = "http://clarin.eu/fcs/1.0/kwic";
    //private static final String FCS_KWIC_PREFIX = "kwic";
        
    private static String CORPUS_LANG = "de";
    private static String QUERY_LANG = "CQL";   
	private Map<String,Corpus> corpusList = new HashMap<String,Corpus>();
    
	private Logger logger = (Logger) LoggerFactory.getLogger(SRU_WS.class);
	private SAXParserFactory factory = SAXParserFactory.newInstance();
			
	private void generateCorpusList() throws IOException{
		corpusList = new HashMap<String, Corpus>();
		
		URI requestURI = null;
		try {
			URIBuilder builder = new URIBuilder(endpointUri);			
			builder.setParameter("operation", "scan");
			builder.setParameter("version", "1.2");
			builder.setParameter("scanClause", "fcs.resource=root");
			requestURI = builder.build();
		} catch (URISyntaxException e) {
			e.printStackTrace();
		}	
		HttpGet request = new HttpGet(requestURI);
		
		CloseableHttpClient httpClient = HttpClients.createDefault();
		CloseableHttpResponse httpResponse = httpClient.execute(request);
		HttpEntity httpEntity = httpResponse.getEntity();
		BufferedInputStream bis = new BufferedInputStream(httpEntity.getContent());			
		
		try {
			SAXParser saxParser = factory.newSAXParser();
			SRUScanResonseHandler handler = new SRUScanResonseHandler();
			saxParser.parse(bis, handler);
			corpusList = handler.getCorpora();
			
		} catch (ParserConfigurationException | SAXException e) {
			e.printStackTrace();
		}
	}
		
	@Override
	protected void doPost(HttpServletRequest req, HttpServletResponse resp)
			throws ServletException, IOException {
						
		String endpoint = req.getParameter("endpoint");
		String query = req.getParameter("query");
		String max = req.getParameter("maximumRecords");
		
		if (endpoint == null || endpoint.isEmpty()){
			resp.sendError(HttpServletResponse.SC_BAD_REQUEST,
					"Endpoint parameter cannot be empty.");
		}
		else if (getServletConfig().getInitParameter(endpointBase+endpoint) == null){
			resp.sendError(HttpServletResponse.SC_BAD_REQUEST,
					"Endpoint parameter is incorrect.");
		}
		else if (query == null || query.isEmpty()){
			resp.sendError(HttpServletResponse.SC_BAD_REQUEST, "Query cannot be empty.");
		}
		else if (max == null || max.isEmpty() || max.equals("0")){
			resp.sendError(HttpServletResponse.SC_BAD_REQUEST, 
					"The maximumRecords must be specified greater than 0.");
		}
		else{			
			
			endpointUri = getServletConfig().getInitParameter(endpointBase+endpoint);
			logger.info("Endpoint URL {} ",endpointUri);		
			
			generateCorpusList();
			
			HttpGet getReq = createSearchRetrieveRequest(query, req.getParameter("maximumRecords"));			
			CloseableHttpClient httpClient = HttpClients.createDefault();
			CloseableHttpResponse httpResponse = httpClient.execute(getReq);
			HttpEntity httpEntity = httpResponse.getEntity();
			BufferedInputStream bis = new BufferedInputStream(httpEntity.getContent());
			
			//List<String[]> tokenizedSentences = new ArrayList<String[]>();
			String text;
			try {
				//text = readInput(bis, tokenizedSentences);
				text = readKwicInput(bis);
				//logger.info(text);
			} catch (XMLStreamException e) {				
				throw new ServletException("XML streaming error");
			}
			
			writeTCL(resp.getOutputStream(), text, query);
			//writeTCL(resp.getOutputStream(), text, tokenizedSentences, query);
		}
	}
	
	public void writeTCL(OutputStream out, String text, List<String[]> tokenizedSentences, String query){
		TextCorpusStored textCorpus = new TextCorpusStored(CORPUS_LANG);
		textCorpus.createTextLayer().addText(text);

		TokensLayer tokensLayer = textCorpus.createTokensLayer();
		SentencesLayer sentencesLayer = textCorpus.createSentencesLayer();	
		MatchesLayer matchesLayer = textCorpus.createMatchesLayer(QUERY_LANG, query);
		MatchedCorpus matchedCorpus = matchesLayer.addCorpus("Wikipedia", "WikiPID");
		
		List<Token> itemToken = new ArrayList<Token>();
		for (int i=0; i<tokenizedSentences.size();i++){			
			List<Token> sentenceTokens = new ArrayList<Token>();				
			for (String tokenString : tokenizedSentences.get(i)) {
				
				Token token = tokensLayer.addToken(tokenString);
				sentenceTokens.add(token);
				
				if (tokenString.equalsIgnoreCase(query)){
					itemToken.add(token);
				}
			}
			sentencesLayer.addSentence(sentenceTokens);
			matchesLayer.addItem(matchedCorpus, itemToken);
			itemToken.clear();
		}
		
		WLData wlData = new WLData(textCorpus);
		try {
			WLDObjector.write(wlData, out);
		} catch (WLFormatException e) {			
			e.printStackTrace();
		}
	}
	
	public void writeTCL(OutputStream out, String text, String query){
		TextCorpusStored textCorpus = new TextCorpusStored(CORPUS_LANG);
		textCorpus.createTextLayer().addText(text);

		TokensLayer tokensLayer = textCorpus.createTokensLayer();
		SentencesLayer sentencesLayer = textCorpus.createSentencesLayer();	
		MatchesLayer matchesLayer = textCorpus.createMatchesLayer(QUERY_LANG, query);
		
		Corpus c;
		MatchedCorpus matchedCorpus; 
		for (String pid : corpusList.keySet()){
			c = corpusList.get(pid);
			if (c.getTokenizedSentences().size() == 0) continue;
			
			matchedCorpus = matchesLayer.addCorpus(c.getName(), c.getPid());
			
			List<String[]> tokenizedSentences = c.getTokenizedSentences();
			List<Token> itemToken = new ArrayList<Token>();			
			
			for (int i=0; i<tokenizedSentences.size();i++){
				
				List<Token> sentenceTokens = new ArrayList<Token>();				
				for (String tokenString : tokenizedSentences.get(i)) {
					
					Token token = tokensLayer.addToken(tokenString);
					sentenceTokens.add(token);
					
					if (tokenString.equalsIgnoreCase(query)){
						itemToken.add(token);
					}
				}
				sentencesLayer.addSentence(sentenceTokens);
				matchesLayer.addItem(matchedCorpus, itemToken);
				itemToken.clear();
			}
		}
		
		WLData wlData = new WLData(textCorpus);
		try {
			WLDObjector.write(wlData, out);
		} catch (WLFormatException e) {			
			e.printStackTrace();
		}
	}
	
	private String readKwicInput(InputStream is) throws XMLStreamException, UnsupportedEncodingException{
		
		String kwic;
		StringBuilder sb = null;
		StringBuilder kwicCollector = new StringBuilder();	
		
		XMLInputFactory f = XMLInputFactory.newInstance();
		XMLStreamReader reader = f.createXMLStreamReader(is);
		
		Corpus c = null;
		int e;
		while (reader.hasNext()) {
			e = reader.next();
			
			if (e == XMLEvent.START_ELEMENT &&
					reader.getLocalName().equals("Resource") &&
					reader.getPrefix().equals(FCS_PREFIX) &&
					reader.getNamespaceURI().equals(FCS_NS)){
				
				for (int i=0; i<reader.getAttributeCount(); i++){
					if (reader.getAttributeLocalName(i).equals("pid")){
						String pid = reader.getAttributeValue(i);						
						logger.info("Corpus: " + URLDecoder.decode(pid,"UTF-8"));						
						c = corpusList.get(URLDecoder.decode(pid,"UTF-8"));
						if (c == null) c = corpusList.get(DUMMY_PID);
						//logger.info(c.getPid());
					}					
				}
			}
				
							
			else if (e == XMLEvent.START_ELEMENT &&
					reader.getLocalName().equals("kwic") ){
				sb = new StringBuilder();
			}
			
			else if (e == XMLEvent.START_ELEMENT &&
					reader.getLocalName().equals("c") ){					
				reader.next();
				sb.append(reader.getText());
				sb.append(" ");
			}				

			else if (e == XMLEvent.START_ELEMENT &&
					reader.getLocalName().equals("kw")){				
				reader.next();
				sb.append(reader.getText());
				sb.append(" ");
			}
			
			else if (e == XMLEvent.START_ELEMENT &&
					reader.getLocalName().equals("c")){					
				reader.next();
				sb.append(reader.getText());
				sb.append(" ");
			}
			
			else if (e == XMLEvent.END_ELEMENT && 
					reader.getLocalName().equals("kwic")){
				kwic = normalizeString(sb.toString());
				//logger.info("sentence {}", kwic);
				c.addSentence(kwic);					
				kwicCollector.append(kwic);
			}				
		}
		return kwicCollector.toString();
	}
	
	private String readInput(InputStream is, List<String[]> tokenizedSentences) 
			throws XMLStreamException{
		
		String kwic;
		StringBuilder sb;
		StringBuilder kwicCollector = new StringBuilder();	
		
		XMLInputFactory f = XMLInputFactory.newInstance();
		XMLStreamReader reader = f.createXMLStreamReader(is);
		
		while (reader.hasNext()) {			
			if (reader.next() == XMLEvent.START_ELEMENT &&
					reader.getLocalName().equals("Result") &&
					reader.getPrefix().equals(HITS_PREFIX) &&
					reader.getNamespaceURI().equals(HITS_NS)){
					
				sb = new StringBuilder();
				
				reader.next();					
				sb.append(reader.getText());
				sb.append(" "); // left context
				
				reader.next();
				if (reader.getLocalName().equals("Hit") &&
						reader.getPrefix().equals(HITS_PREFIX) &&
						reader.getNamespaceURI().equals(HITS_NS)){
					
					reader.next();
					sb.append(reader.getText());
					sb.append(" ");
				}
				reader.next(); // end element Hit					
				reader.next();
				sb.append(reader.getText()); // right context
				
				kwic = normalizeString(sb.toString());
				tokenizedSentences.add(kwic.split("\\s+"));
				kwicCollector.append(kwic);
			}			
		}
		return kwicCollector.toString();
	}
	
	private String normalizeString(String text){
		StringBuilder sb = new StringBuilder();
		for (int j = 0; j < text.length(); j++){
			char temp = text.charAt(j);					
			if (!Character.isDigit(temp) && !Character.isLetter(temp)){						
				if (temp == '\''){ sb.append(temp);	}
				else { sb.append(" "+temp);	}							
			}
			else{ sb.append(temp); }
		}
		return sb.toString();			
	}	
	
	private HttpGet createSearchRetrieveRequest(String query, String maximumRecords) 
			throws ServletException{
		try {
			URIBuilder builder = new URIBuilder(endpointUri);	
			builder.setParameter("operation", "searchRetrieve");
			builder.setParameter("version", "1.2");
			builder.setParameter("query", query);
			builder.setParameter("startRecord", "1");
			builder.setParameter("maximumRecords", maximumRecords);
			URI requestURI = builder.build();
			return new HttpGet(requestURI);
		}
		catch (URISyntaxException e) {
			throw new ServletException();
		}
	}	
}
